#include <netdb.h>
#include <resolv.h>
#include "mep.h"

typedef struct mxlist {
  int pref;
  char *host;
  struct mxlist *next;
} mxlist_t;

void prefinsert(mxlist_t **list, mxlist_t *element)
{
  while (*list != NULL && (*list)->pref < element->pref)
    list = &(*list)->next;

  element->next = *list;
  *list = element;
}

void atsignreplace(char *string)
{
  while (*string != '\0') {
    if (*string == '@')
      *string = '.';

    string++;
  }
}

char *strip(char *string, char delim)
{
  while (*string != '\0') {
    string++;

    if (*string == delim) {
      string++;
      break;
    }
  }

  return string;
}

char *getprefmx(const char *domain)
{
  static mxlist_t *reply = NULL;
  static const char *previous;
  u_char nsreply[NS_PACKETSZ];
  mxlist_t *temp;
  ns_msg handle;
  ns_rr record;
  int len, rnum;

  if (reply != NULL) {
    temp = reply;
    reply = temp->next;
    free(temp->host);
    free(temp);
  }

  if (previous != domain) {
    previous = domain;
    while (reply != NULL) {
      temp = reply;
      reply = temp->next;
      free(temp->host);
      free(temp);
    }

    if ((len = res_search(domain, ns_c_in, ns_t_mx,
			  nsreply, NS_PACKETSZ)) < 0)
      return NULL;

    if (ns_initparse(nsreply, len, &handle) < 0) {
      fprintf(stderr, "ns_initparse: %s\n", strerror(errno));
      return NULL;
    }

    for (rnum = 0; rnum < ns_msg_count(handle, ns_s_an); rnum++) {
      if (ns_parserr(&handle, ns_s_an, rnum, &record) < 0) {
	fprintf(stderr, "ns_parserr: %s\n", strerror(errno));
	return NULL;
      }

      temp = (mxlist_t *) malloc(sizeof(mxlist_t));
      temp->pref = ns_get16(ns_rr_rdata(record));
      temp->host = (char *) malloc(MAXDNAME * sizeof(char));

      if (dn_expand(ns_msg_base(handle), ns_msg_end(handle),
		    ns_rr_rdata(record) + 2, temp->host, MAXDNAME)
	  < 0) {
	fprintf(stderr, "dn_expand: %s\n", strerror(errno));
	return NULL;
      }

      prefinsert(&reply, temp);
    }
  }

  if (reply == NULL)
    return NULL;

  return reply->host;
}

int getmxsocket(char *domain)
{
  char *server;
  int error, sock;
  struct addrinfo hints, *addr;

  if (domain == NULL)
    return 0;

  atsignreplace(domain);

  while (*domain != '\0') {
    if ((server = getprefmx(domain)) != NULL) {
      memset(&hints, 0, sizeof(hints));
      hints.ai_family = AF_INET6;
      hints.ai_socktype = SOCK_STREAM;

      if ((error = getaddrinfo(server, PORT, &hints, &addr)) != 0) {
	fprintf(stderr, "Unable to resolv address: %s\n",
		strerror(errno));
	exit(EXIT_FAILURE);
      }

      if ((sock = socket(PF_INET6, SOCK_STREAM, 0)) < 0) {
	fprintf(stderr, "Unable to get a socket: %s\n",
		strerror(errno));
	exit(EXIT_FAILURE);
      }

      if (connect(sock, addr->ai_addr, sizeof(struct sockaddr_in6))
	  < 0) {
	fprintf(stderr, "Unable to connect: %s\n",
		strerror(errno));
	exit(EXIT_FAILURE);
      }

      return sock;
    }

    domain = strip(domain, '.');
  }

  return 0;
}
